@echo off
rem Copyright (C) 1986, 1987, 1988, 1989, 1990, 1991, 1992, 1993,
rem     1994, 1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003,
rem     2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013,
rem     2014, 2015, 2016, 2017, 2018, 2019, 2020, 2021, 2022
rem     Massachusetts Institute of Technology
copy \scheme\edwin\midas.com \scheme\lib\edwin\autoload\midas.com
copy \scheme\edwin\midas.bci \scheme\lib\edwin\autoload\midas.bci
copy \scheme\edwin\pasmod.com \scheme\lib\edwin\autoload\pasmod.com
copy \scheme\edwin\pasmod.bci \scheme\lib\edwin\autoload\pasmod.bci
copy \scheme\edwin\tximod.com \scheme\lib\edwin\autoload\tximod.com
copy \scheme\edwin\tximod.bci \scheme\lib\edwin\autoload\tximod.bci
copy \scheme\edwin\sort.com \scheme\lib\edwin\autoload\sort.com
copy \scheme\edwin\sort.bci \scheme\lib\edwin\autoload\sort.bci
copy \scheme\edwin\doscom.com \scheme\lib\edwin\autoload\doscom.com
copy \scheme\edwin\doscom.bci \scheme\lib\edwin\autoload\doscom.bci
copy \scheme\edwin\dosshell.com \scheme\lib\edwin\autoload\dosshell.com
copy \scheme\edwin\dosshell.bci \scheme\lib\edwin\autoload\dosshell.bci
copy \scheme\edwin\notify.com \scheme\lib\edwin\autoload\notify.com
copy \scheme\edwin\notify.bci \scheme\lib\edwin\autoload\notify.bci
copy \scheme\edwin\dabbrev.com \scheme\lib\edwin\autoload\dabbrev.com
copy \scheme\edwin\dabbrev.bci \scheme\lib\edwin\autoload\dabbrev.bci
copy \scheme\edwin\info.com \scheme\lib\edwin\autoload\info.com
copy \scheme\edwin\info.bci \scheme\lib\edwin\autoload\info.bci
copy \scheme\edwin\artdebug.com \scheme\lib\edwin\autoload\artdebug.com
copy \scheme\edwin\artdebug.bci \scheme\lib\edwin\autoload\artdebug.bci
