/* limits.h */
/* Produced by hard-params version 4.1, CWI, Amsterdam */

#define CHAR_BIT 8
#define CHAR_MAX 127
#define CHAR_MIN (-128)
#define SCHAR_MAX 127
#define SCHAR_MIN (-128)
#define UCHAR_MAX 255
#define SHRT_MAX 32767
#define SHRT_MIN (-32768)
#define INT_MAX 2147483647
#define INT_MIN (-2147483648)
#define LONG_MAX 2147483647
#define LONG_MIN (-2147483648)
#define USHRT_MAX 65535
#define UINT_MAX 4294967295
#define ULONG_MAX 4294967295
