/* -*-C-*-

Copyright (C) 1986, 1987, 1988, 1989, 1990, 1991, 1992, 1993, 1994,
    1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005,
    2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016,
    2017, 2018, 2019, 2020, 2021, 2022 Massachusetts Institute of
    Technology

This file is part of MIT/GNU Scheme.

MIT/GNU Scheme is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

MIT/GNU Scheme is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with MIT/GNU Scheme; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301,
USA.

*/

/* Interface to the gdbm database library */

#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <gdbm.h>

typedef struct gdbm_args {
  GDBM_FILE dbf;
  gdbm_error errno_gdbm;
  int errno_sys;
  datum key, content;
  int key_allocation, content_allocation;
} gdbm_args;

extern char * alloc_gdbm_key (gdbm_args * args, int size);
extern char * alloc_gdbm_content (gdbm_args * args, int size);

extern gdbm_args * do_gdbm_open (char * name,
				 int block_size, int read_write, int mode);
extern void do_gdbm_close (gdbm_args * args);
extern int do_gdbm_store (gdbm_args * args, int flag);
extern void do_gdbm_fetch (gdbm_args * args);
extern int do_gdbm_exists (gdbm_args * args);
extern int do_gdbm_delete (gdbm_args * args);
extern void do_gdbm_firstkey (gdbm_args * args);
extern int do_gdbm_nextkey (gdbm_args * args);
extern int do_gdbm_reorganize (gdbm_args * args);
extern void do_gdbm_sync (gdbm_args * args);
extern int do_gdbm_setopt (gdbm_args * args, int option, int value);
extern const char * get_gdbm_version (void);
